//
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1FilePreviewViewController.h"

#import <QuickLook/QuickLook.h>

#import "U1ByteSizeValueTransformer.h"
#import "U1FilesClient.h"
#import "U1Node.h"
#import "U1FileNode.h"
#import "U1LocalFileInfo.h"


@interface U1FilePreviewViewController () <QLPreviewControllerDataSource, QLPreviewControllerDelegate>
@property (readwrite, retain) U1FileNode *node;
@property (retain) UIDocumentInteractionController *docController;
@property (copy) NSURL *contentURL;
@property (nonatomic, retain) U1LocalFileInfo *fileInfo;

@property (retain) UIBarButtonItem *publishItem;
@property (retain) UIBarButtonItem *downloadItem;
@property (retain) UIBarButtonItem *previewItem;
@property (retain) UIBarButtonItem *openInItem;
@property (retain) UIBarButtonItem *downloadProgressItem;
@property (retain) UIProgressView *downloadProgressView;

- (UIBarButtonItem*)fixedSpaceItemWithWidth:(CGFloat)width;
- (UIBarButtonItem*)flexibleSpaceItem;

NSDateFormatter *modificationDateFormatter(void);
U1ByteSizeValueTransformer *sizeTransformer(void);

@end


@implementation U1FilePreviewViewController

@synthesize loadingContainerView, iconImageView, nameLabel, loadingActivityView, sizeLabel, createdLabel, modifiedLabel, toolbar, publicURLLabel;
@synthesize node, docController, filesClient, contentURL, fileInfo, publishItem, downloadItem, previewItem, openInItem, downloadProgressItem, downloadProgressView;

NSDateFormatter *modificationDateFormatter(void)
{
	static NSDateFormatter *modificationDateFormatter = nil;
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		modificationDateFormatter = [[NSDateFormatter alloc] init];
		[modificationDateFormatter setDateStyle:NSDateFormatterMediumStyle];
		[modificationDateFormatter setTimeStyle:NSDateFormatterShortStyle];
	});
	return modificationDateFormatter;
}

U1ByteSizeValueTransformer *sizeTransformer(void)
{
	static U1ByteSizeValueTransformer *sizeTransformer = nil;
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		sizeTransformer = [[U1ByteSizeValueTransformer alloc] init];
	});
	return sizeTransformer;
}

- (id)initWithNode:(U1FileNode*)theNode;
{
	self = [super initWithNibName:@"U1FilePreviewViewController" bundle:nil];
	if (self == nil)
		return nil;
	self.node = theNode;
	self.title = [node.path lastPathComponent];
	
	UILabel *titleLabel = [[UILabel alloc] initWithFrame:CGRectMake(0, 0, 100, 20)];
	// TODO: Ubuntu font
	[titleLabel setFont:[UIFont boldSystemFontOfSize:18.f]];
	[titleLabel setTextColor:[UIColor blackColor]];
	[titleLabel setShadowColor:[UIColor whiteColor]];
	[titleLabel setShadowOffset:CGSizeMake(-1, -1)];
	[titleLabel setText:self.title];
	[titleLabel sizeToFit];
	[titleLabel setOpaque:NO];
	[titleLabel setBackgroundColor:[UIColor clearColor]];
	[self.navigationItem setTitleView:titleLabel];
	[titleLabel release];

	publishItem = [[UIBarButtonItem alloc] initWithImage:[UIImage imageNamed:@"ic_act_publish_file"] style:UIBarButtonItemStylePlain target:self action:@selector(publishItem:)];
	downloadItem = [[UIBarButtonItem alloc] initWithImage:[UIImage imageNamed:@"265-download"] style:UIBarButtonItemStylePlain target:self action:@selector(downloadItem:)];
	previewItem = [[UIBarButtonItem alloc] initWithImage:[UIImage imageNamed:@"04-eye"] style:UIBarButtonItemStylePlain target:self action:@selector(showPreview:)];
	openInItem = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemAction target:self action:@selector(openIn:)];
	downloadProgressView = [[UIProgressView alloc] initWithProgressViewStyle:UIProgressViewStyleBar];
	downloadProgressItem = [[UIBarButtonItem alloc] initWithCustomView:downloadProgressView];

	return self;
}

- (void)dealloc;
{
	[loadingContainerView release];
	[iconImageView release];
	[nameLabel release];
	[downloadProgressView release];
	[loadingActivityView release];
	[sizeLabel release];
	[createdLabel release];
	[modifiedLabel release];
	[node release];
	[docController release];
	[contentURL release];
	[fileInfo release];
	[filesClient release];
	[toolbar release];
	[publicURLLabel release];
	[publishItem release];
	[downloadItem release];
	[previewItem release];
	[openInItem release];
	[downloadProgressItem release];
	[super dealloc];
}

- (void)setFileInfo:(U1LocalFileInfo *)value;
{
	if (value == fileInfo)
		return;
	
	[value retain];
	[fileInfo autorelease];
	fileInfo = value;
	self.docController = [UIDocumentInteractionController interactionControllerWithURL:fileInfo.fileURL];
}

- (void)viewDidLoad;
{
	[super viewDidLoad];
	
	[self.nameLabel setText:self.title];
	NSString *sizeString = [sizeTransformer() transformedValue:[self.node size]];
	[self.sizeLabel setText:sizeString];
	NSString *createdString = [modificationDateFormatter() stringFromDate:[self.node creationDate]];
	[self.createdLabel setText:createdString];
	NSString *modifiedString = [modificationDateFormatter() stringFromDate:[self.node modificationDate]];
	[self.modifiedLabel setText:modifiedString];
	[self.publicURLLabel setText:[self.node publicURLString]];
	if ([self.node isPublicValue])
	{
		[self.publishItem setImage:[UIImage imageNamed:@"ic_act_unpublish_file"]];
	}
	else
	{
		[self.publishItem setImage:[UIImage imageNamed:@"ic_act_publish_file"]];
	}
	self.fileInfo = [self.filesClient localInfoForNode:self.node];
	[self.iconImageView setImage:[self.docController.icons lastObject]];
	
	NSArray *items = nil;
	if ([self.fileInfo isDownloaded])
	{
		items = [NSArray arrayWithObjects:
				 [self flexibleSpaceItem],
				 self.publishItem,
				 [self fixedSpaceItemWithWidth:32.f],
				 self.previewItem,
				 [self fixedSpaceItemWithWidth:32.f],
				 self.openInItem,
				 [self flexibleSpaceItem],
				 nil];
	}
	else
	{
		items = [NSArray arrayWithObjects:
				 [self flexibleSpaceItem],
				 self.publishItem,
				 [self fixedSpaceItemWithWidth:32.f],
				 self.downloadItem,
				 [self flexibleSpaceItem],
				 nil];
	}
	[self.toolbar setItems:items animated:NO];
}

- (void)viewDidUnload;
{
	[loadingContainerView release], loadingContainerView = nil;
	[iconImageView release], iconImageView = nil;
	[nameLabel release], nameLabel = nil;
	[loadingActivityView release], loadingActivityView = nil;
	[sizeLabel release], sizeLabel = nil;
	[createdLabel release], createdLabel = nil;
	[modifiedLabel release], modifiedLabel = nil;
	[toolbar release], toolbar = nil;
	[publicURLLabel release], publicURLLabel = nil;
	[super viewDidUnload];
}

- (IBAction)publishItem:(UIBarButtonItem*)sender;
{
	if ([self.node isPublicValue])
	{
		[self.filesClient unpublishNode:self.node completionBlock:^(U1FileNode *node, NSError *error) {
			dispatch_async(dispatch_get_main_queue(), ^(void) {
				[self.publicURLLabel setText:@""];
				[self.publishItem setImage:[UIImage imageNamed:@"ic_act_publish_file"]];
			});
		}];
	}
	else
	{
		[self.filesClient publishNode:self.node completionBlock:^(U1FileNode *node, NSError *error) {
			dispatch_async(dispatch_get_main_queue(), ^(void) {
				[self.publicURLLabel setText:[self.node publicURLString]];
				[self.publishItem setImage:[UIImage imageNamed:@"ic_act_unpublish_file"]];
			});
		}];
	}
}

- (IBAction)downloadItem:(UIBarButtonItem*)sender;
{
	self.fileInfo = [self.filesClient downloadFileForNode:self.node completionBlock:^(NSError *error) {
		[self.fileInfo removeObserver:self forKeyPath:@"downloadPercentage"];
		[self.fileInfo removeObserver:self forKeyPath:@"downloading"];
		[self.fileInfo removeObserver:self forKeyPath:@"downloaded"];
	}];
	[self.fileInfo addObserver:self forKeyPath:@"downloadPercentage" options:0 context:self.fileInfo];
	[self.fileInfo addObserver:self forKeyPath:@"downloading" options:NSKeyValueObservingOptionInitial context:self.fileInfo];
	[self.fileInfo addObserver:self forKeyPath:@"downloaded" options:0 context:self.fileInfo];
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context;
{
	if (context == self.fileInfo && [keyPath isEqualToString:@"downloadPercentage"])
	{
		[self.downloadProgressView setProgress:[self.fileInfo downloadPercentage]];
	}
	else if (context == self.fileInfo && [keyPath isEqualToString:@"downloading"])
	{
		if ([self.fileInfo isDownloading])
		{
			NSArray *downloadingItems = [NSArray arrayWithObjects:
										 [self flexibleSpaceItem],
										 self.publishItem,
										 [self fixedSpaceItemWithWidth:32.f],
										 self.downloadProgressItem,
										 [self flexibleSpaceItem],
										 nil];
			[self.toolbar setItems:downloadingItems animated:YES];
		}
	}
	else if (context == self.fileInfo && [keyPath isEqualToString:@"downloaded"])
	{
		if ([self.fileInfo isDownloaded])
		{
			NSArray *downloadedItems = [NSArray arrayWithObjects:
										[self flexibleSpaceItem],
										self.publishItem,
										[self fixedSpaceItemWithWidth:32.f],
										self.previewItem,
										[self fixedSpaceItemWithWidth:32.f],
										self.openInItem,
										[self flexibleSpaceItem],
										nil];
			[self.toolbar setItems:downloadedItems animated:YES];
		}
	}
	else
	{
		[super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
	}
}

- (IBAction)showPreview:(UIBarButtonItem*)sender;
{
	QLPreviewController *previewController = [[QLPreviewController alloc] init];
	previewController.dataSource = self;
	previewController.delegate = self;

	UILabel *titleLabel = [[UILabel alloc] initWithFrame:CGRectMake(0, 0, 100, 20)];
	// TODO: Ubuntu font
	[titleLabel setFont:[UIFont boldSystemFontOfSize:18.f]];
	[titleLabel setTextColor:[UIColor blackColor]];
	[titleLabel setShadowColor:[UIColor whiteColor]];
	[titleLabel setShadowOffset:CGSizeMake(-1, -1)];
	[titleLabel setText:self.title];
	[titleLabel sizeToFit];
	[titleLabel setOpaque:NO];
	[titleLabel setBackgroundColor:[UIColor clearColor]];
	[previewController.navigationItem setTitleView:titleLabel];
	[titleLabel release];

//	[self.navigationController presentModalViewController:previewController animated:YES];
	[self.navigationController pushViewController:previewController animated:YES];
	[previewController release];

	// TODO: manage preview view vs. modal presentation
//	UIView *previewView = previewController.view;
//	previewView.frame = [self.view bounds];
//	[previewController viewWillAppear:NO];
//	[self.view addSubview:previewView];
//	[previewController viewDidAppear:NO];
}

- (IBAction)openIn:(UIBarButtonItem*)sender;
{
	[self.docController presentOpenInMenuFromBarButtonItem:sender animated:YES];
}

- (NSInteger) numberOfPreviewItemsInPreviewController: (QLPreviewController *) controller;
{
	return 1;
}

- (id <QLPreviewItem>) previewController: (QLPreviewController *) controller previewItemAtIndex: (NSInteger) index;
{
	return [self.fileInfo fileURL];
}

- (UIBarButtonItem*)fixedSpaceItemWithWidth:(CGFloat)width;
{
	UIBarButtonItem *item = [[[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemFixedSpace target:nil action:NULL] autorelease];
	item.width = width;
	return item;
}

- (UIBarButtonItem*)flexibleSpaceItem;
{
	return [[[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemFlexibleSpace target:nil action:NULL] autorelease];
}

@end
