//
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1FilesClient.h"

#import <CoreData/CoreData.h>

#import "U1DataRepository.h"
#import "U1FilesService.h"
#import "U1FileNode.h"
#import "U1FolderNode.h"
#import "U1LocalFileInfo.h"
#import "U1Node.h"


@interface U1FilesClient ()
- (NSURL*)fileURLForNode:(U1FileNode*)node;
@end


@implementation U1FilesClient

@synthesize dataRepository, filesService;

- (void)dealloc;
{
	[dataRepository release];
	[filesService release];
	[super dealloc];
}

- (id)fetchNodeWithResourcePath:(NSString*)nodePath completionBlock:(void(^)(U1Node *node, NSFetchedResultsController *childrenResultsController, NSError *error))completionBlock;
{
	
	U1FolderNode *node = (id)[self.dataRepository nodeWithResourcePath:nodePath];
	
	// TODO: temporary, create node if doesn't exist
	if (!node)
	{
		[self.dataRepository dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {
			U1Node *newNode = [U1FolderNode insertInManagedObjectContext:context];
			newNode.resourcePath = nodePath;
			[context save:NULL];
		}];
	}

	node = (id)[self.dataRepository nodeWithResourcePath:nodePath];
	
	[self.filesService infoForNodeAtResourcePath:node.resourcePath includeChildren:YES completionBlock:^(NSDictionary *nodeInfo, NSError *error) {
		
		[node updatePropertiesFromJSONDictionary:nodeInfo];
		
		NSArray *childInfos = [nodeInfo objectForKey:@"children"];
		[self.dataRepository dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {

			for (NSDictionary *childInfo in childInfos)
			{
				NSString *childResourcePath = [childInfo objectForKey:@"resource_path"];
				NSString *kind = [childInfo objectForKey:@"kind"];
				U1Node *childNode = [self.dataRepository nodeWithResourcePath:childResourcePath];
				if (!childNode)
				{
					if ([@"file" isEqualToString:kind])
						childNode = [U1FileNode insertInManagedObjectContext:context];
					else
						childNode = [U1FolderNode insertInManagedObjectContext:context];
				}
				
				[childNode updatePropertiesFromJSONDictionary:childInfo];
				childNode.parent = node;
			}
			[context save:NULL];
		}];
		
		__block NSFetchedResultsController *resultsController = nil;
		[self.dataRepository dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {
			
			NSFetchRequest *request = [[NSFetchRequest alloc] init];
			[request setEntity:[U1Node entityInManagedObjectContext:context]];
			NSPredicate *p = [NSPredicate predicateWithFormat:@"parent = %@", node];
			[request setPredicate:p];
			NSArray *sortDescriptors = [NSArray arrayWithObjects:
										[NSSortDescriptor sortDescriptorWithKey:@"kind" ascending:YES],
										[NSSortDescriptor sortDescriptorWithKey:@"resourcePath" ascending:YES selector:@selector(caseInsensitiveCompare:)],
										nil];
			[request setSortDescriptors:sortDescriptors];
			resultsController = [[NSFetchedResultsController alloc] initWithFetchRequest:request managedObjectContext:context sectionNameKeyPath:nil cacheName:nil];
			[request release];
		}];
		
		
		dispatch_async(dispatch_get_main_queue(), ^(void) {
			NSError *error = nil;
			[resultsController performFetch:&error];
			completionBlock(node, resultsController, error);
			[resultsController release];
		});
		
	}];
	return nil;
}

- (id)deleteNode:(U1Node*)node completionBlock:(void(^)(NSError *error))completionBlock;
{
	return [self.filesService deleteNodeAtResourcePath:node.resourcePath completionBlock:^(NSError *error) {
		if (!error)
		{
			[self.dataRepository dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {
				[context deleteObject:node];
				[context save:NULL];
			}];
		}
		completionBlock(error);
	}];
}

- (id)uploadContentData:(NSData*)contentData toFolder:(U1FolderNode*)folderNode withResourceName:(NSString*)resourceName progressBlock:(void(^)(long long bytesUploaded, long long totalBytes))progressBlock completionBlock:(void(^)(U1FileNode *updatedNode, NSError *error))completionBlock;
{
	NSString *resourcePath = [folderNode.resourcePath stringByAppendingPathComponent:resourceName];
	U1FileNode *node = (id)[self.dataRepository nodeWithResourcePath:resourcePath];
	if (!node)
	{
		[self.dataRepository dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {
			U1FileNode *newNode = [U1FileNode insertInManagedObjectContext:context];
			newNode.resourcePath = resourcePath;
			newNode.kind = @"file";
			newNode.contentPath = [folderNode.contentPath stringByAppendingPathComponent:resourceName];
			newNode.parent = folderNode;
			[context save:NULL];
		}];
		node = (id)[self.dataRepository nodeWithResourcePath:resourcePath];
	}
	
	return [self.filesService uploadContentData:contentData forNode:node progressBlock:progressBlock completionBlock:^(NSDictionary *updatedNodeInfo, NSError *error) {
		
		[node updatePropertiesFromJSONDictionary:updatedNodeInfo];
		completionBlock(node, error);
	}];
}

- (id)publishNode:(U1FileNode*)node completionBlock:(void(^)(U1FileNode *node, NSError *error))completionBlock;
{
	return [self.filesService publishNode:node completionBlock:completionBlock];
}

- (id)unpublishNode:(U1FileNode*)node completionBlock:(void(^)(U1FileNode *node, NSError *error))completionBlock;
{
	return [self.filesService unpublishNode:node completionBlock:completionBlock];
}

- (BOOL)isNodeSavedLocally:(U1FileNode*)node;
{
	NSString *contentPath = node.contentPath;
	
	NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
	NSString *documentsDirectory = [paths objectAtIndex:0];
	
	NSString *filePath = [documentsDirectory stringByAppendingPathComponent:contentPath];
	NSFileManager *fileManager = [[[NSFileManager alloc] init] autorelease];
	return [fileManager fileExistsAtPath:filePath];
}

- (U1LocalFileInfo*)localInfoForNode:(U1FileNode*)node;
{
	// TODO: cache info
	U1LocalFileInfo *info = [[U1LocalFileInfo new] autorelease];
	info.fileURL = [self fileURLForNode:node];
	NSFileManager *fileManager = [[[NSFileManager alloc] init] autorelease];
	info.downloaded = [fileManager fileExistsAtPath:[info.fileURL path]];
	info.downloading = NO;
	return info;
}

- (NSURL*)fileURLForNode:(U1FileNode*)node;
{
	NSString *contentPath = node.contentPath;
	NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
	NSString *documentsDirectory = [paths objectAtIndex:0];
	NSString *filePath = [documentsDirectory stringByAppendingPathComponent:contentPath];
	NSURL *fileURL = [NSURL fileURLWithPath:filePath];
	return fileURL;
}

- (U1LocalFileInfo*)downloadFileForNode:(U1FileNode*)node completionBlock:(void(^)(NSError *error))completionBlock;
{
	U1LocalFileInfo *info = [[U1LocalFileInfo new] autorelease];
	info.fileURL = [self fileURLForNode:node];
	info.downloaded = NO;
	info.downloading = YES;
	info.downloadPercentage = 0.f;

	CGFloat expected = [[node size] longLongValue];
	
	[self.filesService contentForNode:node progressBlock:^(long long bytesReceived, long long bytesExpected) {
		
		dispatch_async(dispatch_get_main_queue(), ^(void) {

			CGFloat received = (CGFloat)bytesReceived;
			info.downloadPercentage = (received / expected);
			
		});

	} completionBlock:^(NSURL *contentURL) {
		dispatch_async(dispatch_get_main_queue(), ^(void) {
			
			info.downloadPercentage = 100.f;
			info.downloading = NO;
			info.downloaded = YES;
			
		});
	}];

	return info;
}

@end
