//
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1FilesService.h"

#import "JSONKit.h"
#import "OAuthConsumer.h"

#import "U1AccountManager.h"
#import "U1HTTPRequestOperation.h"
#import "U1Node.h"
#import "U1FileNode.h"


@interface U1FilesService ()

@property (retain) NSOperationQueue *networkQueue;

- (NSURL*)URLForNode:(NSString*)nodePath includeChildren:(BOOL)includeChildren;
- (id)execute:(NSString*)method toURL:(NSURL*)url withParameters:(NSDictionary*)parameters requestBody:(NSString*)body parseResponseBody:(BOOL)parseResponseBody completionBlock:(void(^)(id results, NSError *error))completionBlock;
- (id)execute:(NSString*)method toURL:(NSURL*)url withParameters:(NSDictionary*)parameters requestData:(NSData*)bodyData contentType:(NSString*)contentType parseResponseBody:(BOOL)parseResponseBody completionBlock:(void(^)(id results, NSError *error))completionBlock;

@end


@implementation U1FilesService

static NSString *const U1FilesServiceAPIBase = @"https://one.ubuntu.com";
static NSString *const U1FilesServiceAPIPath = @"/api/file_storage";
static NSString *const U1FilesServiceAPIVersion1 = @"/v1";
static NSString *const U1FilesServiceAPIContentRoot = @"https://files.one.ubuntu.com";

@synthesize networkQueue;

- (id)init;
{
	self = [super init];
	if (self == nil)
		return nil;
	networkQueue = [[NSOperationQueue alloc] init];
	return self;
}

- (void)dealloc;
{
	[networkQueue release];
	[super dealloc];
}

- (id)volumeInfoWithCompletionBlock:(void(^)(NSArray *volumeInfo, NSError *error))completionBlock;
{
	NSURL *volumesURL = [self URLForNode:@"/volumes" includeChildren:NO];
	return [self execute:@"GET" toURL:volumesURL withParameters:nil requestBody:nil parseResponseBody:YES completionBlock:^(id results, NSError *error) {
		completionBlock(results, error);
	}];
}

- (id)infoForNodeAtResourcePath:(NSString*)resourcePath includeChildren:(BOOL)includeChildren completionBlock:(void(^)(NSDictionary *node, NSError *error))completionBlock;
{
	NSURL *nodeURL = [self URLForNode:resourcePath includeChildren:includeChildren];
	return [self execute:@"GET" toURL:nodeURL withParameters:nil requestBody:nil parseResponseBody:YES completionBlock:^(id result, NSError *error) {
		
		completionBlock(result, error);
	}];
}

- (id)infoForNode:(U1Node*)node includeChildren:(BOOL)includeChildren completionBlock:(void(^)(NSDictionary *node, NSError *error))completionBlock;
{
	return [self infoForNodeAtResourcePath:node.resourcePath includeChildren:includeChildren completionBlock:completionBlock];
}

- (id)contentForNode:(U1FileNode*)node progressBlock:(void(^)(long long bytesReceived, long long bytesExpected))progressBlock completionBlock:(void(^)(NSURL *contentURL))completionBlock;
{
	NSString *contentPath = node.contentPath;
	
	NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
	NSString *documentsDirectory = [paths objectAtIndex:0];
	
	NSString *filePath = [documentsDirectory stringByAppendingPathComponent:contentPath];
	NSFileManager *fileManager = [[[NSFileManager alloc] init] autorelease];
	NSString *directory = [filePath stringByDeletingLastPathComponent];
	if (![fileManager fileExistsAtPath:directory isDirectory:NULL])
	{
		[fileManager createDirectoryAtPath:directory withIntermediateDirectories:YES attributes:nil error:NULL];
	}

	NSURL *fileURL = [NSURL fileURLWithPath:filePath];
	if ([fileManager fileExistsAtPath:filePath])
	{
		NSError *error = nil;
		NSDictionary *attributes = [fileManager attributesOfItemAtPath:filePath error:&error];
		
		// if the file doesn't seem to be modified, return from documents directory
		if ([[node size] unsignedLongLongValue] == [attributes fileSize] &&
			[[node modificationDate] compare:[attributes fileModificationDate]] == NSOrderedAscending)
		{
			completionBlock(fileURL);
			return nil;
		}
	}

	NSURL *requestURL = [NSURL URLWithString:[[U1FilesServiceAPIContentRoot stringByAppendingPathComponent:contentPath] stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]];
	OAToken *token = [[U1AccountManager sharedAccountManager] oauthToken];
	OAConsumer *consumer = [[U1AccountManager sharedAccountManager] oauthConsumer];
	id<OASignatureProviding> signatureProvider = [OAPlaintextSignatureProvider new];
	OAMutableURLRequest *request = [[OAMutableURLRequest alloc] initWithURL:requestURL consumer:consumer token:token realm:nil signatureProvider:signatureProvider];
	[signatureProvider release];
	[request prepare];

	U1HTTPRequestOperationFileDataCollector *fileCollector = [[U1HTTPRequestOperationFileDataCollector alloc] initWithURL:fileURL];
	U1HTTPRequestOperationReportingDataCollector *collector = [[U1HTTPRequestOperationReportingDataCollector alloc] initWithDataCollector:fileCollector];
	[fileCollector release];
	collector.progressBlock = progressBlock;
	U1HTTPRequestOperation *httpOperation = [[U1HTTPRequestOperation alloc] initWithRequest:request dataCollector:collector];
	[httpOperation setCompletionBlock:^(void) {
		[collector release];
		if (completionBlock)
			completionBlock(fileURL);
	}];
	[self.networkQueue addOperation:httpOperation];
	return nil;
}

- (id)uploadContentAtURL:(NSURL*)contentURL forNode:(U1FileNode*)node progressBlock:(void(^)(long long bytesUploaded, long long totalBytes))progressBlock completionBlock:(void(^)(U1FileNode *updatedNode, NSError *error))completionBlock;
{
	NSString *contentPath = node.contentPath;
	NSURL *requestURL = [NSURL URLWithString:[[U1FilesServiceAPIContentRoot stringByAppendingPathComponent:contentPath] stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]];
	OAToken *token = [[U1AccountManager sharedAccountManager] oauthToken];
	OAConsumer *consumer = [[U1AccountManager sharedAccountManager] oauthConsumer];
	id<OASignatureProviding> signatureProvider = [OAPlaintextSignatureProvider new];
	OAMutableURLRequest *request = [[OAMutableURLRequest alloc] initWithURL:requestURL consumer:consumer token:token realm:nil signatureProvider:signatureProvider];
	[signatureProvider release];
	
	[request setHTTPMethod:@"PUT"];
	[request setValue:@"application/octet-stream" forHTTPHeaderField:@"Content-Type"];
	[request prepare];
	NSInputStream *contentStream = [[NSInputStream alloc] initWithURL:contentURL];
	[request setHTTPBodyStream:contentStream];
	
	U1HTTPRequestOperationInMemoryDataCollector *collector = [U1HTTPRequestOperationInMemoryDataCollector new];
	U1HTTPRequestOperation *httpOperation = [[U1HTTPRequestOperation alloc] initWithRequest:request dataCollector:collector];
	[request release];
	[httpOperation setCompletionBlock:^(void) {
		[collector release];
//		if (completionBlock)
//			completionBlock(fileURL);
	}];
	[self.networkQueue addOperation:httpOperation];

	return nil;
}

- (id)uploadContentData:(NSData*)contentData forNode:(U1FileNode*)node progressBlock:(void(^)(long long bytesUploaded, long long totalBytes))progressBlock completionBlock:(void(^)(NSDictionary *updatedNodeInfo, NSError *error))completionBlock;
{
	NSString *contentPath = node.contentPath;
	NSURL *requestURL = [NSURL URLWithString:[[U1FilesServiceAPIContentRoot stringByAppendingPathComponent:contentPath] stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]];
	
	[self execute:@"PUT" toURL:requestURL withParameters:nil requestData:contentData contentType:@"application/octet-stream" parseResponseBody:YES completionBlock:^(id results, NSError *error) {
		
		completionBlock(results, error);
	}];
	
	return nil;
}

- (id)deleteNodeAtResourcePath:(NSString*)resourcePath completionBlock:(void(^)(NSError *error))completionBlock;
{
	NSURL *nodeURL = [self URLForNode:resourcePath includeChildren:NO];
	return [self execute:@"DELETE" toURL:nodeURL withParameters:nil requestBody:nil parseResponseBody:NO completionBlock:^(id result, NSError *error) {
		completionBlock(error);
	}];
}

- (id)publishNode:(U1FileNode*)node completionBlock:(void(^)(U1FileNode *node, NSError *error))completionBlock;
{
	NSString *nodePath = node.resourcePath;
	NSURL *nodeURL = [self URLForNode:(NSString*)nodePath includeChildren:NO];
	return [self execute:@"PUT" toURL:nodeURL withParameters:nil requestBody:@"{\"is_public\": true}" parseResponseBody:YES completionBlock:^(id result, NSError *error) {

		[node updatePropertiesFromJSONDictionary:result];
		completionBlock(node, error);
	}];
}

- (id)unpublishNode:(U1FileNode*)node completionBlock:(void(^)(U1FileNode *node, NSError *error))completionBlock;
{
	NSString *nodePath = node.resourcePath;
	NSURL *nodeURL = [self URLForNode:(NSString*)nodePath includeChildren:NO];
	return [self execute:@"PUT" toURL:nodeURL withParameters:nil requestBody:@"{\"is_public\": false}" parseResponseBody:YES completionBlock:^(id result, NSError *error) {

		[node updatePropertiesFromJSONDictionary:result];
		completionBlock(node, error);
	}];
}


#pragma mark Private Methods

- (NSURL*)URLForNode:(NSString*)nodePath includeChildren:(BOOL)includeChildren;
{
	NSMutableString *urlString = [NSMutableString stringWithString:U1FilesServiceAPIBase];
	[urlString appendString:U1FilesServiceAPIPath];
	[urlString appendString:U1FilesServiceAPIVersion1];
	[urlString appendString:nodePath];
	if (includeChildren)
		[urlString appendString:@"?include_children=true"];
	NSString *finalString = [urlString stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding];
	return [NSURL URLWithString:finalString];
}

- (id)execute:(NSString*)method toURL:(NSURL*)url withParameters:(NSDictionary*)parameters requestBody:(NSString*)body parseResponseBody:(BOOL)parseResponseBody completionBlock:(void(^)(id results, NSError *error))completionBlock;
{
	return [self execute:method toURL:url withParameters:parameters requestData:[body dataUsingEncoding:NSUTF8StringEncoding] contentType:@"application/json" parseResponseBody:parseResponseBody completionBlock:completionBlock];
}

- (id)execute:(NSString*)method toURL:(NSURL*)url withParameters:(NSDictionary*)parameters requestData:(NSData*)bodyData contentType:(NSString*)contentType parseResponseBody:(BOOL)parseResponseBody completionBlock:(void(^)(id results, NSError *error))completionBlock;
{

	OAToken *token = [[U1AccountManager sharedAccountManager] oauthToken];
	OAConsumer *consumer = [[U1AccountManager sharedAccountManager] oauthConsumer];
	id<OASignatureProviding> signatureProvider = [OAPlaintextSignatureProvider new];
	OAMutableURLRequest *request = [[OAMutableURLRequest alloc] initWithURL:url consumer:consumer token:token realm:nil signatureProvider:signatureProvider];
	[signatureProvider release];
	[request setHTTPMethod:method];
	if (bodyData)
	{
		[request setHTTPBody:bodyData];
		[request setValue:contentType forHTTPHeaderField:@"Content-Type"];
	}
	[request prepare];
	
	U1HTTPRequestOperationInMemoryDataCollector *collector = [U1HTTPRequestOperationInMemoryDataCollector new];
	U1HTTPRequestOperation *httpOperation = [[U1HTTPRequestOperation alloc] initWithRequest:request dataCollector:collector];
	[request release];
	
	[httpOperation setCompletionBlock:^(void) {
		dispatch_async(dispatch_get_main_queue(), ^(void) {
			
			id result = nil;
			NSHTTPURLResponse *response = httpOperation.response;
			NSError *error = httpOperation.error;
			if (!error &&
				![[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(200, 100)] containsIndex:[response statusCode]])
			{
				// TODO: semantic error messages
				error = [NSError errorWithDomain:@"U1FilesServiceErrorDomain" code:0 userInfo:nil];
			}
			if (!error && parseResponseBody)
			{
				NSData *data = [collector data];
				result = [data objectFromJSONDataWithParseOptions:JKParseOptionNone error:&error];
			}
			
			if (completionBlock)
				completionBlock(result, error);
			
			[collector release];
		});
	}];
	[self.networkQueue addOperation:httpOperation];
	[httpOperation release];
	
	return nil;
}

@end
